﻿using System;
using System.Linq;
using System.Windows.Forms;
using System.Collections.Generic;
using System.ComponentModel;

/*                  Backpropagation with the Iris data set                       */
/*                                                                               */
/*                       www.mosismath.com                                       */

namespace WindowsFormsApplication1
{
   
    public partial class Mainwin : Form
    {

        int samples = 29;
        int dataLength = 0;
        const int iterations = 200000; //60000;
        double learning_rate = 0.6; // 0.4; // reduce to 0.2 for tanh
        const int features = 4;
        int layers = 1;
        double cost = 1000;
        double costSum = 0;
 
        TData data = new TData();
        List<TLayer> net = new List<TLayer>();
        BackgroundWorker calculationThread;


        public Mainwin()
        {
            int i,j;
            DataGridViewCell cell;
            TLayer tempRow;

            InitializeComponent();
            data.readData(Application.StartupPath + "\\data.csv");
             
            tempRow = new TLayer(4, 2);
            tempRow.Act = tempRow.Relu;
            tempRow.dAct = tempRow.dRelu;
            net.Add(tempRow);
            tempRow = new TLayer(2, 2);
            net.Add(tempRow);
           
            layers = net.Count;

            gRise.ColumnCount = 4;
            gRise.RowCount = 10;
            samples = data.values.Count;
            dataLength = data.dataLength;
            dgDataView.RowCount = samples;
            dgDataView.ColumnCount = 10;

            dgDataView.Columns[0].HeaderText = "x1";
            dgDataView.Columns[1].HeaderText = "x2";
            dgDataView.Columns[2].HeaderText = "x3";
            dgDataView.Columns[3].HeaderText = "x4";
            dgDataView.Columns[4].HeaderText = "y";
            dgDataView.Columns[5].HeaderText = "out 1";
            dgDataView.Columns[6].HeaderText = "out 2";

            for (i = 0; i < samples; i++)
            {
                dgDataView.Rows[i].HeaderCell.Value = (i + 1).ToString();
                for (j = 0; j <= features; j++)
                {
                    cell = dgDataView[j, i];
                    cell.Value = Math.Round(data.values.ElementAt(i)[j], 4).ToString();
                }
            }

            calculationThread = new BackgroundWorker();
            calculationThread.DoWork += calculate;
            calculationThread.RunWorkerCompleted += calculationCompleted;
        }


        private double[] ForwardProp(double[] x)
        {
            int i, j, k;
            TLayer actLayer = net.ElementAt(0);
            double[] actX = new double[x.Length];
            for (i = 0; i < x.Length; i++)
              actX[i] = x[i];

            for (i = 0; i < layers; i++)
            {
                for (j = 0; j < actLayer.featuresOut; j++)
                {
                    actLayer.x[j] = actLayer.offs[j];
                    for (k = 0; k < actLayer.featuresIn; k++)
                    {
                        actLayer.x[j] = actLayer.x[j] + actX[k] * actLayer.w[k, j];
                    }
                    actLayer.o[j] = actLayer.Act(actLayer.x[j]);
                }

                for (j = 0; j < actLayer.x.Length; j++)
                    actX[j] = actLayer.o[j];
              
                if(i < layers - 1)
                    actLayer = net.ElementAt(i + 1);
            }
            return actLayer.o;
        }


        private void BackwardProp(double[] x, double[] y)
        {
            int i, j, k;
            TLayer actLayer = net.ElementAt(layers - 2);   // actual computed layer
            TLayer layerRight = net.ElementAt(layers - 1); // nex layer to the right of the actual one
            double[] actX = new double[x.Length];
            for (j = 0; j < actLayer.x.Length; j++)
                actX[j] = actLayer.x[j];
            actLayer = net.ElementAt(layers - 1);
            for (j = 0; j < actLayer.featuresOut; j++)
            {
                costSum = costSum + ((y[j] - actLayer.o[j]) * (y[j] - actLayer.o[j]));
                actLayer.gradient[j] = (y[j] - actLayer.o[j]) * -actLayer.dAct(actLayer.o[j]);
            }

            for (j = 0; j < actLayer.featuresIn; j++)
            {
                for (k = 0; k < actLayer.featuresOut; k++)
                {
                    actLayer.deltaW[j, k] = actLayer.deltaW[j, k] + actLayer.gradient[k] * actX[k];
                }
            }

            for (j = 0; j < actLayer.featuresOut; j++)
            {
                actLayer.deltaOffs[j] = actLayer.deltaOffs[j] + actLayer.gradient[j];
            }
          
            if (layers > 1)
            {
                for (i = layers - 2; i >= 0; i--)
                {
                    actLayer = net.ElementAt(i);
                    layerRight = net.ElementAt(i + 1);
                    if (i > 0)
                    {
                        TLayer layerLeft = net.ElementAt(i - 1);
                        for (j = 0; j < layerLeft.o.Length; j++)
                            actX[j] = layerLeft.o[j];
                    }
                    else
                    {
                        for (j = 0; j < x.Length; j++)
                            actX[j] = x[j];
                    }
                    for (j = 0; j < actLayer.featuresOut; j++)
                    {
                        actLayer.gradient[j] = 0;
                        for (k = 0; k < layerRight.featuresOut; k++)
                        {
                            actLayer.gradient[j] = actLayer.gradient[j] + (layerRight.gradient[k] * layerRight.w[j, k]);
                        }
                        actLayer.gradient[j] = actLayer.gradient[j] * actLayer.dAct(actLayer.o[j]);
                        actLayer.deltaOffs[j] = actLayer.deltaOffs[j] + actLayer.gradient[j];
                        for (k = 0; k < actLayer.featuresIn; k++)
                        {
                            actLayer.deltaW[k, j] = actLayer.deltaW[k, j] + actLayer.gradient[j] * actX[k];
                        }
                    }
                }
            }
        }


        private void UpdateValues()
        {
            int i, k, l;
            TLayer actLayer = net.ElementAt(layers - 1);

            for (i = 0; i < layers; i++)
            {
                actLayer = net.ElementAt(i);
                for (k = 0; k < actLayer.featuresOut; k++)
                {
                    for (l = 0; l < actLayer.featuresIn; l++)
                    {
                        actLayer.w[l, k] = actLayer.w[l, k] - actLayer.deltaW[l, k] / samples * learning_rate;
                    }
                    actLayer.offs[k] = actLayer.offs[k] - actLayer.deltaOffs[k] / samples * learning_rate;
                }
            }
        }


        private void calculate(object sender, System.ComponentModel.DoWorkEventArgs e)
        {
            int i, j, k;
            int count = 0;
            double[] tempX = new double[features];
            double[] tempY = new double[2];
            TLayer actLayer = net.ElementAt(0);

            while (count < iterations)
            {
                actLayer = net.ElementAt(0);

                for (i = 0; i < layers; i++)
                {
                    actLayer = net.ElementAt(i);
                    for (j = 0; j < actLayer.featuresOut; j++)
                    {
                        actLayer.gradient[j] = 0;
                        actLayer.deltaOffs[j] = 0;
                        for (k = 0; k < actLayer.featuresIn; k++)
                        {
                            actLayer.deltaW[k, j] = 0;
                        }
                    }
                }
                count++;

                costSum = 0;
                for (i = 0; i < samples; i++)
                {

                    if (data.values.ElementAt(i)[features] == 1)
                    {
                        tempY[0] = 0;
                        tempY[1] = 0;
                    }
                    else
                    {
                        if (data.values.ElementAt(i)[features] == 2)
                        {
                            tempY[0] = 1;
                            tempY[1] = 0;
                        }
                        else
                        {
                            tempY[0] = 1;
                            tempY[1] = 1;
                        }
                    }
                    for (k = 0; k < features; k++)
                    {
                        tempX[k] = data.values.ElementAt(i)[k];
                    }
                    ForwardProp(tempX);
                    BackwardProp(tempX, tempY);
                }
                UpdateValues();
                cost = costSum / samples;
            }
        }

        void calculationCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            int i, j, k;
            double[] res;
            double[] tempX = new double[features];
            DataGridViewCell cell;
            TLayer actLayer = net.ElementAt(0);
            tbCost.Text = Math.Round(cost, 8).ToString();

            // print data
            for (i = 0; i < samples; i++)
            {
                for (k = 0; k < features; k++)
                {
                    tempX[k] = data.values.ElementAt(i)[k];
                }
                res = ForwardProp(tempX);
                cell = dgDataView[5, i];
                cell.Value = Math.Round(res[0], 4).ToString();
                cell = dgDataView[6, i];
                cell.Value = Math.Round(res[1], 4).ToString();
            }

            actLayer = net.ElementAt(0);
            int rowCount = 0;
            for (i = 0; i < layers; i++)
            {
                actLayer = net.ElementAt(i);
                for (j = 0; j < actLayer.featuresIn; j++)
                {
                    for (k = 0; k < actLayer.featuresOut; k++)
                    {
                        cell = gRise[k, rowCount + j];
                        cell.Value = Math.Round(actLayer.w[j, k], 6).ToString();
                    }
                    if (j < actLayer.offs.Length)
                    {
                        cell = gRise[k + 1, rowCount + j];
                        cell.Value = Math.Round(actLayer.offs[j], 6).ToString();
                    }
                }
                rowCount = rowCount + actLayer.featuresIn + 1;
            }
            Cursor.Current = Cursors.Default;
            this.Cursor = Cursors.Default;
            dgDataView.Cursor = Cursors.Default;
            gRise.Cursor = Cursors.Default;
            bt_Calc.Enabled = true;
        }

        private void bt_Calc_Click(object sender, EventArgs e)
        {
            this.Cursor = Cursors.WaitCursor;
            Cursor.Current = Cursors.WaitCursor;
            bt_Calc.Enabled = false;
            calculationThread.RunWorkerAsync();

        }
    }
}
